/* start of file PolRoots.c --------------------------------------------- */

/* <summary>
	SUBROUTINE PolRoots - Computes the n roots on the complex polynomial
	p = p[0], p[1],..., p[n]
   </summary>
*/

#include <stdbool.h>	// for bool
#include "PolRoots.h"

void PolNorm(double _Complex a[], int *n);
void Laguer(double _Complex a[], int n, int *its, double _Complex *x);
void ClearSmall(double _Complex x);

void PolRoots(double _Complex a[], int n, bool polish, double _Complex roots[])
/* roots = The m = n-1 roots of complex polynomial a = a[0], a[1],..., a[m]
*/
/* Adapted from http://www.library.cornell.edu/nr/bookfpdf/f9-5.pdf
   The coefficient a[m] is the constant term, while a[0] is the
   coefficient of the highest power of x. This routine calls Laguer in
   succession for each root, performs the deflation, optionally
   polishes the roots, if polish = true, by the same Laguerre method
   -- if you are not going to polish in some other way -- and finally
   sorts the roots by their real parts.
*/
{
	int m = n - 1;			/* n = size of polynomial, m = degree of polynomial */
	int i, j, jj, its;
	double _Complex aa[maxN];
	double _Complex cZero = 0.0 + I * 0.0;
	for (i = 0; i < n; i++)	/* aa = a */
		aa[i] = a[i];
	int aan = n;
	PolNorm(aa, &aan);
	double _Complex ad[maxN];
	double _Complex x;
	double _Complex b;
	double _Complex c;
	for (i = 0; i < n; i++)	/* ad = aa */
		ad[i] = aa[i];
	int adn = n;
	for (j = 0; j < m; j++)		/* Loop over each root to be found. */
	{
		/* Start at zero to favor convergence to smallest remaining root */
		x = cZero;
		Laguer(ad, adn, &its, &x);	/* Find the root */
		ClearSmall(x);				/* clear small real or imaginary part */
		roots[j] = x;				/* save the root */
		b = ad[0];					/* Forward deflation. */
		for (jj = 1; jj <= m - j; jj++)
		{
			c = ad[jj];
			ad[jj - 1] = b;
			b = x * b + c;
		}
		adn--;	/* fix ad to have one less row */
	}
	if (polish)
	{
		/* Polish the roots using the undeflated coefficients */
		for (j = 0; j < m; j++)
			Laguer(a, n, &its, &roots[j]);
	}
	for (j = 1; j < m; j++)	/* Sort roots by straight insertion */
	{
		x = roots[j];
		for (i = j - 1; i >= 0; i--)
		{

			if (creal(roots[i]) < creal(x))	/* if (roots[i] <= x) */
				break;
			if (creal(roots[i]) == creal(x) && cimag(roots[i]) <= cimag(x))
				break;
			roots[i + 1] = roots[i];
		}
		roots[i + 1] = x;
	}
} /* PolRoots */

void PolNorm(double _Complex a[], int *n)
/* a = norm of polynomial a, leading coeff. not zero
*/
{
	if (creal(a[0]) != 0.0 || cimag(a[0]) != 0.0)
		return;
	int nz = 0;
	int r;
	for (r = 0; r < *n - 1; r++)
		if (creal(a[r]) == 0.0 && cimag(a[r]) == 0.0)
			nz++;
		else
			break;
	for (r = nz; r < *n; r++)
		a[r - nz] = a[r];
	*n -= nz;
} /* PolNorm */

void ClearSmall(double _Complex x)
{
	if (creal(x) != 0.0 && cimag(x) / creal(x) <= 9.3E-16)
	{
		x = creal(x) + I * 0.0;
		return;
	}
	if (cimag(x) != 0.0 && creal(x) / cimag(x) <= 9.3E-16)
		x = 0.0 + I * cimag(x);
} /* ClearSmall */

/* end of file PolRoots.c ----------------------------------------------- */

